@@ -132,7 +132,8 @@ impl<T: 'static> Drop for Arc<T> {
     }
 }
 
-unsafe impl<T: 'static> Send for Arc<T> {}
+unsafe impl<T: 'static + Send> Send for Arc<T> {}
+unsafe impl<T: 'static + Sync> Sync for Arc<T> {}
 
 #[cfg(test)]
 mod test {
@@ -141,15 +142,17 @@ mod test {
     use std::sync::atomic::Ordering::Relaxed;
     use std::sync::atomic::{AtomicBool, AtomicUsize};
 
+    struct A(AtomicUsize, usize, &'static AtomicBool);
+    impl Drop for A {
+        fn drop(&mut self) {
+            self.2.swap(true, Relaxed);
+        }
+    }
+
     #[test]
     fn arc() {
         static DESTROYED: AtomicBool = AtomicBool::new(false);
-        struct A(AtomicUsize, usize, &'static AtomicBool);
-        impl Drop for A {
-            fn drop(&mut self) {
-                self.2.swap(true, Relaxed);
-            }
-        }
+
         let mut arc = Arc::new(A(AtomicUsize::new(10), 10, &DESTROYED));
         if let Some(mut_ref) = arc.get_mut() {
             mut_ref.1 += 1;
@@ -182,4 +185,30 @@ mod test {
             drop(Barrier::new());
         }
     }
+
+    #[test]
+    fn arc_send() {
+        static DESTROYED: AtomicBool = AtomicBool::new(false);
+
+        let arc = Arc::new(A(AtomicUsize::new(14), 14, &DESTROYED));
+        let arc_cloned = arc.clone();
+        let thread = std::thread::spawn(move || {
+            assert_eq!(arc_cloned.0.load(Relaxed), arc_cloned.1);
+        });
+        assert!(thread.join().is_ok());
+        assert_eq!(arc.0.load(Relaxed), arc.1);
+    }
+
+    #[test]
+    fn arc_arc_send() {
+        static DESTROYED: AtomicBool = AtomicBool::new(false);
+
+        let arc_arc = std::sync::Arc::new(Arc::new(A(AtomicUsize::new(14), 14, &DESTROYED)));
+        let arc_arc_cloned = arc_arc.clone();
+        let thread = std::thread::spawn(move || {
+            assert_eq!(arc_arc_cloned.0.load(Relaxed), 14);
+        });
+        assert!(thread.join().is_ok());
+        assert_eq!(arc_arc.0.load(Relaxed), 14);
+    }
 }
