@@ -2,12 +2,18 @@
 
 //! Scalable concurrent containers.
 //!
-//! # [`ebr`]
+//! # [`EBR`](ebr)
 //!
-//! The [`ebr`] module implements epoch-based reclamation for [`HashMap`], [`HashIndex`], and [`TreeIndex`].
+//! The [`ebr`] module implements epoch-based reclamation for [`LinkedList`], [`HashMap`],
+//! [`HashIndex`], and [`TreeIndex`].
+//!
+//! # [`LinkedList`]
+//! [`LinkedList`] is a type trait that implements wait-free list modification operations for
+//! a generic concurrent list.
 //!
 //! # [`HashMap`]
-//! [`HashMap`] is a concurrent hash map that dynamically grows and shrinks without blocking other operations.
+//! [`HashMap`] is a concurrent hash map that dynamically grows and shrinks without blocking
+//! other operations.
 //!
 //! # [`HashIndex`]
 //! [`HashIndex`] is a read-optimized concurrent hash index that is similar to [`HashMap`].
@@ -17,6 +23,9 @@
 
 pub mod ebr;
 
+mod linked_list;
+pub use linked_list::LinkedList;
+
 pub mod hash_map;
 pub use hash_map::HashMap;
 
