@@ -16,7 +16,7 @@ use std::fmt;
 use std::iter::FusedIterator;
 use std::ops::Bound::{Excluded, Included, Unbounded};
 use std::ops::RangeBounds;
-use std::sync::atomic::Ordering::{AcqRel, Acquire};
+use std::sync::atomic::Ordering::{AcqRel, Acquire, Relaxed};
 
 /// A scalable concurrent B+ tree.
 ///
@@ -152,6 +152,28 @@ where
     /// ```
     #[inline]
     pub fn remove<Q>(&self, key_ref: &Q) -> bool
+    where
+        K: Borrow<Q>,
+        Q: Ord + ?Sized,
+    {
+        self.remove_if(key_ref, |_| true)
+    }
+
+    /// Removes a key-value pair if the given condition is met.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use scc::TreeIndex;
+    ///
+    /// let treeindex: TreeIndex<u64, u32> = TreeIndex::new();
+    ///
+    /// assert!(treeindex.insert(1, 10).is_ok());
+    /// assert!(!treeindex.remove_if(&1, |v| *v == 0));
+    /// assert!(treeindex.remove_if(&1, |v| *v == 10));
+    /// ```
+    #[inline]
+    pub fn remove_if<Q, F: FnMut(&V) -> bool>(&self, key_ref: &Q, mut condition: F) -> bool
     where
         K: Borrow<Q>,
         Q: Ord + ?Sized,
@@ -160,14 +182,14 @@ where
         let barrier = Barrier::new();
         let mut root_ptr = self.root.load(Acquire, &barrier);
         while let Some(root_ref) = root_ptr.as_ref() {
-            match root_ref.remove(key_ref, &barrier) {
+            match root_ref.remove_if(key_ref, &mut condition, &barrier) {
                 Ok(removed) => return removed || has_been_removed,
                 Err(remove_error) => match remove_error {
                     RemoveError::Empty(removed) => {
                         if removed && !has_been_removed {
                             has_been_removed = true;
                         }
-                        if Node::remove_root(&self.root, true, &barrier) {
+                        if Node::remove_root(&self.root, &barrier) {
                             return has_been_removed;
                         }
                     }
@@ -181,7 +203,7 @@ where
             };
             root_ptr = self.root.load(Acquire, &barrier);
         }
-        false
+        has_been_removed
     }
 
     /// Reads a key-value pair.
@@ -245,7 +267,7 @@ where
     /// ```
     #[inline]
     pub fn clear(&self) {
-        Node::remove_root(&self.root, false, &Barrier::new());
+        self.root.swap((None, Tag::None), Relaxed);
     }
 
     /// Returns the size of the [`TreeIndex`].
@@ -405,16 +427,6 @@ where
     }
 }
 
-impl<K, V> Drop for TreeIndex<K, V>
-where
-    K: 'static + Clone + Ord + Send + Sync,
-    V: 'static + Clone + Send + Sync,
-{
-    fn drop(&mut self) {
-        Node::remove_root(&self.root, false, &Barrier::new());
-    }
-}
-
 /// [`Visitor`] scans all the key-value pairs in the [`TreeIndex`].
 ///
 /// It is guaranteed to visit all the key-value pairs that outlive the [`Visitor`], and it
@@ -452,13 +464,16 @@ where
     fn next(&mut self) -> Option<Self::Item> {
         // Starts scanning.
         if self.leaf_scanner.is_none() {
-            let root_ptr = self.tree.root.load(Acquire, self.barrier);
-            if let Some(root_ref) = root_ptr.as_ref() {
-                if let Ok(scanner) = root_ref.min(self.barrier) {
-                    self.leaf_scanner.replace(scanner);
+            loop {
+                let root_ptr = self.tree.root.load(Acquire, self.barrier);
+                if let Some(root_ref) = root_ptr.as_ref() {
+                    if let Ok(scanner) = root_ref.min(self.barrier) {
+                        self.leaf_scanner.replace(scanner);
+                        break;
+                    }
+                } else {
+                    return None;
                 }
-            } else {
-                return None;
             }
         }
 
