@@ -1,5 +1,5 @@
-use crate::common::linked_list::LinkedList;
-use crate::ebr::{Arc, AtomicArc, Barrier, Tag};
+use crate::ebr::{Arc, AtomicArc, Barrier};
+use crate::LinkedList;
 
 use std::borrow::Borrow;
 use std::cmp::Ordering;
@@ -46,14 +46,7 @@ where
     entry_array: EntryArray<K, V>,
 
     /// A pointer that points to the next adjacent leaf.
-    ///
-    /// forward_link being tagged 1 means that the next leaf may contain smaller keys.
-    forward_link: AtomicArc<Leaf<K, V>>,
-
-    /// A pointer that points to the previous adjacent leaf.
-    ///
-    /// backward_link being tagged 1 means the leaf is locked.
-    backward_link: AtomicArc<Leaf<K, V>>,
+    link: AtomicArc<Leaf<K, V>>,
 
     /// The metadata that manages the contents.
     metadata: AtomicU32,
@@ -68,8 +61,7 @@ where
     pub fn new() -> Leaf<K, V> {
         Leaf {
             entry_array: unsafe { MaybeUninit::uninit().assume_init() },
-            forward_link: AtomicArc::null(),
-            backward_link: AtomicArc::null(),
+            link: AtomicArc::null(),
             metadata: AtomicU32::new(0),
         }
     }
@@ -184,12 +176,16 @@ where
         }
     }
 
-    /// Removes the key.
+    /// Removes the key if the condition is met.
     ///
     /// The first boolean value returned from the function indicates that an entry has been removed.
     /// The second boolean value indicates that the leaf is full.
     /// The third boolean value indicates that the leaf is empty.
-    pub fn remove<Q>(&self, key: &Q) -> (bool, bool, bool)
+    pub fn remove_if<Q, F: FnMut(&V) -> bool>(
+        &self,
+        key: &Q,
+        condition: &mut F,
+    ) -> (bool, bool, bool)
     where
         K: Borrow<Q>,
         Q: Ord + ?Sized,
@@ -216,6 +212,11 @@ where
                     }
                     Ordering::Equal => {
                         loop {
+                            if !condition(self.read(i).1) {
+                                // The given condition is not met.
+                                break;
+                            }
+
                             let new_metadata =
                                 (metadata & (!Self::rank_mask(i))) | Self::rank_bits(i, REMOVED);
                             match self.metadata.compare_exchange(
@@ -377,8 +378,7 @@ where
 
     /// Compares the given metadata value with the current one.
     pub fn validate(&self, metadata: u32) -> bool {
-        // The acquire fence ensures that a reader having read the latest state must read the updated metadata.
-        std::sync::atomic::fence(Acquire);
+        // `Relaxed` is sufficient as long as the caller has read-acquired its contents.
         self.metadata.load(Relaxed) == metadata
     }
 
@@ -747,18 +747,14 @@ where
     }
 }
 
-/// `LinkedList` implementation for `Leaf`.
+/// [`LinkedList`] implementation for [`Leaf`].
 impl<K, V> LinkedList for Leaf<K, V>
 where
     K: 'static + Clone + Ord + Sync,
     V: 'static + Clone + Sync,
 {
-    fn forward_link(&self) -> &AtomicArc<Leaf<K, V>> {
-        &self.forward_link
-    }
-
-    fn backward_link(&self) -> &AtomicArc<Leaf<K, V>> {
-        &self.backward_link
+    fn link_ref(&self) -> &AtomicArc<Leaf<K, V>> {
+        &self.link
     }
 }
 
@@ -864,44 +860,28 @@ where
         min_allowed_key: Option<&K>,
         barrier: &'b Barrier,
     ) -> Option<Scanner<'b, K, V>> {
-        let mut next_ptr = self.leaf.forward_link().load(Acquire, barrier);
-        let mut being_split = next_ptr.tag() == Tag::First;
-        while let Some(next_ref) = next_ptr.as_ref() {
-            let mut leaf_scanner = Scanner::new(next_ref);
-            if let Some(key) = min_allowed_key.as_ref() {
-                if being_split
-                    || leaf_scanner
-                        .leaf
-                        .backward_link()
-                        .load(Acquire, barrier)
-                        .as_raw()
-                        != self.leaf as *const _
-                {
-                    // Data race resolution: compare keys if the current leaf has been unlinked.
+        let mut next_leaf_ptr = self.leaf.next_ptr(Acquire, barrier);
+        while let Some(next_leaf_ref) = next_leaf_ptr.as_ref() {
+            let mut leaf_scanner = Scanner::new(next_leaf_ref);
+            if let Some(key) = min_allowed_key {
+                if !self.leaf.is_clear(Relaxed) {
+                    // Data race resolution: compare keys if the current leaf has been deleted.
                     //
-                    // There is a chance that the current leaf has been unlinked,
-                    // and smaller keys have been inserted into the next leaf.
-                    //  - Next leaf = nl, current leaf = l, leaf node = ln, unlink = u.
-                    //  - Remove: u_back(nl)|u_forward(l)|release|update(ln)|
-                    //  - Insert:           |                               |insert(nl)|
-                    //  - Scan:   load(nl)  |                               |          |load(nl)
-                    // Therefore, it reads the backward link of the next leaf if it points to the current leaf.
-                    // If not, it compares keys in order not to return keys smaller than the current one.
-                    //  - Remove: u_back(nl)|u_forward(l)|release|update(ln)|
-                    //  - Insert:           |                               |insert(nl)|
-                    //  - Scan:   load(nl)  |                               |          |load(nl)|acquire|validate(nl->l)
+                    // There is a chance that the current leaf has been deleted, and smaller
+                    // keys have been inserted into the next leaf.
                     while let Some(entry) = leaf_scanner.next() {
-                        if key.cmp(&entry.0) == Ordering::Less {
+                        if key.cmp(entry.0) == Ordering::Less {
                             return Some(leaf_scanner);
                         }
                     }
+                    next_leaf_ptr = next_leaf_ref.next_ptr(Acquire, barrier);
+                    continue;
                 }
             }
             if leaf_scanner.next().is_some() {
                 return Some(leaf_scanner);
             }
-            next_ptr = leaf_scanner.leaf.forward_link().load(Acquire, barrier);
-            being_split = being_split || next_ptr.tag() == Tag::First;
+            next_leaf_ptr = next_leaf_ref.next_ptr(Acquire, barrier);
         }
         None
     }
@@ -943,14 +923,14 @@ mod test {
         assert_eq!(leaf.max(), Some((&50, &51)));
         assert!(leaf.insert(60, 61).is_none());
         assert!(leaf.insert(70, 71).is_none());
-        assert!(leaf.remove(&60).0);
+        assert!(leaf.remove_if(&60, &mut |v| *v == 61).0);
         assert!(leaf.insert(60, 61).is_none());
-        assert_eq!(leaf.remove(&60), (true, false, false));
+        assert_eq!(leaf.remove_if(&60, &mut |v| *v == 61), (true, false, false));
         assert!(!leaf.full());
         assert!(leaf.insert(40, 40).is_none());
         assert!(leaf.insert(30, 31).is_none());
         assert!(!leaf.full());
-        assert_eq!(leaf.remove(&40), (true, false, false));
+        assert_eq!(leaf.remove_if(&40, &mut |_| true), (true, false, false));
         assert!(leaf.insert(40, 41).is_none());
         assert_eq!(leaf.insert(30, 33), Some(((30, 33), true)));
         assert!(leaf.insert(10, 11).is_none());
@@ -1032,15 +1012,16 @@ mod test {
         assert!(leaf.insert(13, 14).is_none());
         assert_eq!(*leaf.search(&13).unwrap(), 14);
         assert_eq!(leaf.insert(13, 14), Some(((13, 14), true)));
-        assert_eq!(leaf.remove(&10), (true, false, false));
-        assert_eq!(leaf.remove(&11), (true, false, false));
-        assert_eq!(leaf.remove(&12), (true, false, false));
+        assert_eq!(leaf.remove_if(&10, &mut |v| *v == 11), (true, false, false));
+        assert_eq!(leaf.remove_if(&11, &mut |v| *v == 0), (false, false, false));
+        assert_eq!(leaf.remove_if(&11, &mut |v| *v == 12), (true, false, false));
+        assert_eq!(leaf.remove_if(&12, &mut |v| *v == 13), (true, false, false));
         assert!(!leaf.full());
-        assert!(leaf.remove(&20).0);
+        assert!(leaf.remove_if(&20, &mut |_| true).0);
         assert!(leaf.insert(20, 21).is_none());
         assert!(leaf.search(&11).is_none());
-        assert_eq!(leaf.remove(&10), (false, true, false));
-        assert_eq!(leaf.remove(&11), (false, true, false));
+        assert_eq!(leaf.remove_if(&10, &mut |_| true), (false, true, false));
+        assert_eq!(leaf.remove_if(&11, &mut |_| true), (false, true, false));
         assert_eq!(*leaf.search(&20).unwrap(), 21);
         assert_eq!(leaf.max(), Some((&20, &21)));
         assert!(leaf.full());
@@ -1079,11 +1060,14 @@ mod test {
             assert!(leaf.insert(i, i).is_none());
         }
         for i in 0..ARRAY_SIZE - 1 {
-            assert_eq!(leaf.remove(&i), (true, true, false));
+            assert_eq!(leaf.remove_if(&i, &mut |v| *v == i), (true, true, false));
         }
         assert!(leaf.full());
         assert!(!leaf.obsolete());
-        assert_eq!(leaf.remove(&(ARRAY_SIZE - 1)), (true, true, true));
+        assert_eq!(
+            leaf.remove_if(&(ARRAY_SIZE - 1), &mut |_| true),
+            (true, true, true)
+        );
         assert_eq!(
             leaf.insert(ARRAY_SIZE, ARRAY_SIZE),
             Some(((ARRAY_SIZE, ARRAY_SIZE), false))
@@ -1107,7 +1091,10 @@ mod test {
         for i in 0..ARRAY_SIZE - 2 {
             assert!(leaf1.insert(i, i).is_none());
             assert!(leaf2.insert(i + 4, i + 4).is_none());
-            assert_eq!(leaf2.remove(&(i + 4)), (true, false, true));
+            assert_eq!(
+                leaf2.remove_if(&(i + 4), &mut |_| true),
+                (true, false, true)
+            );
         }
         assert!(!leaf1.retire());
         assert!(!leaf1.obsolete());
@@ -1141,7 +1128,7 @@ mod test {
                     if result.is_none() {
                         assert_eq!(*leaf_copied.search(&tid).unwrap(), 1);
                         if tid % 2 != 0 {
-                            assert!(leaf_copied.remove(&tid).0);
+                            assert!(leaf_copied.remove_if(&tid, &mut |v| *v == 1).0);
                         }
                     }
                     let mut scanner = Scanner::new(&leaf_copied);
