@@ -2,8 +2,8 @@ use super::leaf::{Scanner, ARRAY_SIZE};
 use super::Leaf;
 use super::{InsertError, RemoveError, SearchError};
 
-use crate::common::linked_list::LinkedList;
 use crate::ebr::{Arc, AtomicArc, Barrier, Ptr, Tag};
+use crate::LinkedList;
 
 use std::borrow::Borrow;
 use std::cmp::Ordering;
@@ -86,7 +86,7 @@ where
                 // `child_ptr` being null indicates that the leaf node is bound to be freed.
                 return Err(SearchError::Retry);
             }
-            let unbounded_ptr = (self.leaves.1).load(Relaxed, barrier);
+            let unbounded_ptr = (self.leaves.1).load(Acquire, barrier);
             if let Some(unbounded_ref) = unbounded_ptr.as_ref() {
                 debug_assert!(unbounded_ptr.tag() == Tag::None);
                 if !(self.leaves.0).validate(result.1) {
@@ -121,7 +121,7 @@ where
                 // `child_ptr` being null indicates that the leaf node is bound to be freed.
                 return Err(SearchError::Retry);
             }
-            let unbounded_ptr = (self.leaves.1).load(Relaxed, barrier);
+            let unbounded_ptr = (self.leaves.1).load(Acquire, barrier);
             if let Some(unbounded_ref) = unbounded_ptr.as_ref() {
                 debug_assert!(unbounded_ptr.tag() == Tag::None);
                 if !(self.leaves.0).validate(metadata) {
@@ -160,7 +160,7 @@ where
                 // `child_ptr` being null indicates that the leaf node is bound to be freed.
                 return Err(SearchError::Retry);
             }
-            let unbounded_ptr = (self.leaves.1).load(Relaxed, barrier);
+            let unbounded_ptr = (self.leaves.1).load(Acquire, barrier);
             if let Some(unbounded_ref) = unbounded_ptr.as_ref() {
                 debug_assert!(unbounded_ptr.tag() == Tag::None);
                 if !(self.leaves.0).validate(metadata) {
@@ -208,7 +208,7 @@ where
                 return Err(InsertError::Retry((key, value)));
             }
 
-            let mut unbounded_ptr = self.leaves.1.load(Relaxed, barrier);
+            let mut unbounded_ptr = self.leaves.1.load(Acquire, barrier);
             while unbounded_ptr.is_null() {
                 if unbounded_ptr.tag() == Tag::First {
                     // The leaf node has become obsolete.
@@ -220,8 +220,8 @@ where
                 match self.leaves.1.compare_exchange(
                     unbounded_ptr,
                     (Some(Arc::new(Leaf::new())), Tag::None),
-                    Release,
-                    Relaxed,
+                    AcqRel,
+                    Acquire,
                 ) {
                     Ok((_, ptr)) => {
                         unbounded_ptr = ptr;
@@ -261,7 +261,12 @@ where
     }
 
     /// Removes an entry associated with the given key.
-    pub fn remove<Q>(&self, key_ref: &Q, barrier: &Barrier) -> Result<bool, RemoveError>
+    pub fn remove_if<Q, F: FnMut(&V) -> bool>(
+        &self,
+        key_ref: &Q,
+        condition: &mut F,
+        barrier: &Barrier,
+    ) -> Result<bool, RemoveError>
     where
         K: Borrow<Q>,
         Q: Ord + ?Sized,
@@ -275,10 +280,8 @@ where
                     continue;
                 }
                 if let Some(child_ref) = child_ptr.as_ref() {
-                    let (removed, full, empty) = child_ref.remove(key_ref);
-                    if !full && !empty {
-                        return Ok(removed);
-                    } else if !empty {
+                    let (removed, full, empty) = child_ref.remove_if(key_ref, condition);
+                    if full && !self.check_full_leaf(key_ref, child_ptr, barrier) {
                         // Data race resolution.
                         //  - Insert: start to insert into a full leaf
                         //  - Remove: start removing an entry from the leaf after pointer validation
@@ -286,31 +289,39 @@ where
                         //  - Remove: find the leaf full, and the leaf node is not locked, returning 'Ok(true)'
                         // Consequently, the key remains.
                         // In order to resolve this, check the pointer again.
-                        return self.check_full_leaf(removed, key_ref, child_ptr, barrier);
+                        return Err(RemoveError::Retry(removed));
                     }
-                    child_ref.retire();
-                    return self.coalesce(removed, barrier);
+                    if empty {
+                        if !child_ref.retire() {
+                            return Err(RemoveError::Retry(removed));
+                        }
+                        return self.coalesce(removed, barrier);
+                    }
+                    return Ok(removed);
                 }
 
                 // `child_ptr` being null indicates that the leaf node is bound to be freed.
                 return Err(RemoveError::Retry(false));
             }
-            let unbounded_ptr = (self.leaves.1).load(Relaxed, barrier);
+            let unbounded_ptr = (self.leaves.1).load(Acquire, barrier);
             if let Some(unbounded_ref) = unbounded_ptr.as_ref() {
                 debug_assert!(unbounded_ptr.tag() == Tag::None);
                 if !(self.leaves.0).validate(result.1) {
                     // Data race resolution - see LeafNode::search.
                     continue;
                 }
-                let (removed, full, empty) = unbounded_ref.remove(key_ref);
-                if !full && !empty {
-                    return Ok(removed);
-                } else if !empty {
+                let (removed, full, empty) = unbounded_ref.remove_if(key_ref, condition);
+                if full && !self.check_full_leaf(key_ref, unbounded_ptr, barrier) {
                     // Data race resolution - see above.
-                    return self.check_full_leaf(removed, key_ref, unbounded_ptr, barrier);
+                    return Err(RemoveError::Retry(removed));
                 }
-                unbounded_ref.retire();
-                return self.coalesce(removed, barrier);
+                if empty {
+                    if !unbounded_ref.retire() {
+                        return Err(RemoveError::Retry(removed));
+                    }
+                    return self.coalesce(removed, barrier);
+                }
+                return Ok(removed);
             }
             if unbounded_ptr.tag() == Tag::First {
                 // The leaf node has become obsolete.
@@ -389,26 +400,42 @@ where
         let high_key_leaf_ptr = new_leaves_ref.high_key_leaf.load(Relaxed, barrier);
         let unused_leaf = if high_key_leaf_ptr.is_null() {
             // From here, Scanners can reach the new leaf.
-            full_leaf_ref.push_back(full_leaf_ptr, &[low_key_leaf_ptr], Tag::First, barrier);
+            let result = full_leaf_ref.push_back(
+                low_key_leaf_ptr.try_into_arc().unwrap(),
+                true,
+                Release,
+                barrier,
+            );
+            debug_assert!(result.is_ok());
             // Replaces the full leaf with the low-key leaf.
             full_leaf.swap((low_key_leaf_ptr.try_into_arc(), Tag::None), Release)
         } else {
             // From here, Scanners can reach the new leaves.
             //
-            // Immediately unlinking the full leaf causes the current scanners reading the full leaf
+            // Immediately unlinking the full leaf causes active scanners reading the full leaf
             // to omit a number of leaves.
             //  - Leaf: l, insert: i, split: s, rollback: r.
             //  - Insert 1-2: i1(l1)|                   |i2(l2)|s(l2)|l12:l2:l21:l22|l12:l21:l22|
             //  - Insert 0  : i0(l1)|s(l1)|l1:l11:l12:l2|                                       |r|l1:l12...
-            // In this scenario, without keeping l1, l2 in the linked list, l1 would point to l2,
-            // and therefore a range scanner would start from l1, and cannot traverse l21 and l22,
-            // missing newly inserted entries in l21 and l22 before starting the range scanner.
-            full_leaf_ref.push_back(
-                full_leaf_ptr,
-                &[low_key_leaf_ptr, high_key_leaf_ptr],
-                Tag::First,
+            // In this scenario, without keeping l1, l2 in the linked list, l1 would point to
+            // l2, and therefore a range scanner would start from l1, and cannot traverse l21
+            // and l22, missing newly inserted entries in l21 and l22 before starting the
+            // range scanner.
+            let result = full_leaf_ref.push_back(
+                high_key_leaf_ptr.try_into_arc().unwrap(),
+                true,
+                Release,
                 barrier,
             );
+            debug_assert!(result.is_ok());
+            let result = full_leaf_ref.push_back(
+                low_key_leaf_ptr.try_into_arc().unwrap(),
+                true,
+                Release,
+                barrier,
+            );
+            debug_assert!(result.is_ok());
+
             // Takes the max key value stored in the low key leaf as the leaf key.
             let max_key = low_key_leaf_ptr.as_ref().unwrap().max().unwrap().0;
             if self
@@ -431,7 +458,8 @@ where
         // Drops the deprecated leaf.
         if let Some(unused_leaf) = unused_leaf {
             debug_assert!(unused_leaf.full());
-            unused_leaf.pop_self(Tag::None, barrier);
+            let deleted = unused_leaf.delete_self(Release);
+            debug_assert!(deleted);
             barrier.reclaim(unused_leaf);
         }
 
@@ -555,13 +583,7 @@ where
     }
 
     /// Checks the given full leaf whether it is being split.
-    fn check_full_leaf<Q>(
-        &self,
-        removed: bool,
-        key_ref: &Q,
-        leaf_ptr: Ptr<Leaf<K, V>>,
-        barrier: &Barrier,
-    ) -> Result<bool, RemoveError>
+    fn check_full_leaf<Q>(&self, key_ref: &Q, leaf_ptr: Ptr<Leaf<K, V>>, barrier: &Barrier) -> bool
     where
         K: Borrow<Q>,
         Q: Ord + ?Sized,
@@ -578,7 +600,7 @@ where
             .compare_exchange(Ptr::null(), (None, Tag::None), AcqRel, Relaxed)
             .is_err()
         {
-            return Err(RemoveError::Retry(removed));
+            return false;
         }
         let result = (self.leaves.0).min_greater_equal(key_ref);
         let leaf_current_ptr = if let Some((_, child)) = result.0 {
@@ -586,11 +608,7 @@ where
         } else {
             self.leaves.1.load(Relaxed, barrier)
         };
-        if leaf_current_ptr == leaf_ptr {
-            Ok(removed)
-        } else {
-            Err(RemoveError::Retry(removed))
-        }
+        leaf_current_ptr == leaf_ptr
     }
 
     /// Tries to coalesce empty or obsolete leaves.
@@ -600,23 +618,26 @@ where
             return Err(RemoveError::Retry(removed));
         }
 
-        let mut empty = false;
+        let mut num_valid_leaves = 0;
         for entry in Scanner::new(&self.leaves.0) {
             let leaf_ptr = entry.1.load(Relaxed, barrier);
             let leaf_ref = leaf_ptr.as_ref().unwrap();
             if leaf_ref.obsolete() {
                 // Data race resolution - see LeafScanner::jump.
-                leaf_ref.pop_self(Tag::None, barrier);
-                empty = self.leaves.0.remove(entry.0).2;
+                let deleted = leaf_ref.delete_self(Relaxed);
+                debug_assert!(deleted);
+                self.leaves.0.remove_if(entry.0, &mut |_| true);
                 // Data race resolution - see LeafNode::search.
                 if let Some(leaf) = entry.1.swap((None, Tag::None), Release) {
                     barrier.reclaim(leaf);
                 }
+            } else {
+                num_valid_leaves += 1;
             }
         }
 
         // Checks the unbounded leaf only when all the other leaves have become obsolete.
-        let check_unbounded = if empty && self.leaves.0.retire() {
+        let check_unbounded = if num_valid_leaves == 0 && self.leaves.0.retire() {
             true
         } else {
             self.leaves.0.obsolete()
@@ -628,7 +649,8 @@ where
             if let Some(unbounded_ref) = unbounded_ptr.as_ref() {
                 if unbounded_ref.obsolete() {
                     // Data race resolution - see LeafScanner::jump.
-                    unbounded_ref.pop_self(Tag::None, barrier);
+                    let deleted = unbounded_ref.delete_self(Relaxed);
+                    debug_assert!(deleted);
                     if let Some(obsolete_leaf) = self.leaves.1.swap((None, Tag::First), Release) {
                         barrier.reclaim(obsolete_leaf);
                     }
@@ -661,12 +683,27 @@ where
 
             // Rolls back the linked list state.
             //
-            // In order to keep the forward_link flag of the origin leaf,
-            // high_key_leaf must be opted out first.
-            if let Some(leaf_ref) = high_key_leaf_ptr
-                .as_ref() { leaf_ref.pop_self(Tag::None, barrier); }
-            if let Some(leaf_ref) = low_key_leaf_ptr
-                .as_ref() { leaf_ref.pop_self(Tag::None, barrier); }
+            // `high_key_leaf` must be deleted first in order for scanners not to omit entries.
+            if let Some(leaf_ref) = high_key_leaf_ptr.as_ref() {
+                let deleted = leaf_ref.delete_self(Relaxed);
+                debug_assert!(deleted);
+            }
+            if let Some(leaf_ref) = low_key_leaf_ptr.as_ref() {
+                let deleted = leaf_ref.delete_self(Release);
+                debug_assert!(deleted);
+            }
+
+            if let Some(origin_leaf) = new_leaves_ref
+                .origin_leaf_ptr
+                .swap((None, Tag::None), Relaxed)
+            {
+                // Remove marks from the full leaf node.
+                //
+                // This unmarking has to be a release-store, otherwise it can be re-ordered
+                // before previous `delete_self` calls.
+                let unmarked = origin_leaf.unmark(Release);
+                debug_assert!(unmarked);
+            }
 
             // Unlocks the leaf node.
             if let Some(new_leaves) = self.new_leaves.swap((None, Tag::None), Release) {
@@ -692,7 +729,8 @@ where
                 .swap((None, Tag::None), Relaxed)
             {
                 // Makes the leaf unreachable before dropping it.
-                obsolete_leaf.pop_self(Tag::None, barrier);
+                let deleted = obsolete_leaf.delete_self(Relaxed);
+                debug_assert!(deleted);
                 barrier.reclaim(obsolete_leaf);
             }
         }
@@ -708,8 +746,8 @@ impl<K: Clone + Display + Ord + Send + Sync, V: Clone + Display + Send + Sync> L
     ) -> std::io::Result<()> {
         // Collects information.
         #[allow(clippy::type_complexity)]
-        let mut leaf_ref_array: [Option<(Option<&Leaf<K, V>>, Option<&K>, usize)>; ARRAY_SIZE + 1] =
-            [None; ARRAY_SIZE + 1];
+        let mut leaf_ref_array: [Option<(Option<&Leaf<K, V>>, Option<&K>, usize)>;
+            ARRAY_SIZE + 1] = [None; ARRAY_SIZE + 1];
         let mut scanner = Scanner::new_including_removed(&self.leaves.0);
         let mut index = 0;
         while let Some(entry) = scanner.next() {
@@ -784,14 +822,6 @@ impl<K: Clone + Display + Ord + Send + Sync, V: Clone + Display + Send + Sync> L
                     ))?;
                 }
                 output.write_fmt(format_args!("</table>\n>]\n"))?;
-                let prev_ptr = leaf_ref.backward_link().load(Acquire, barrier);
-                if let Some(prev_ref) = prev_ptr.as_ref() {
-                    output.write_fmt(format_args!("{} -> {}\n", leaf_ref.id(), prev_ref.id()))?;
-                }
-                let next_ptr = leaf_ref.forward_link().load(Acquire, barrier);
-                if let Some(next_ref) = next_ptr.as_ref() {
-                    output.write_fmt(format_args!("{} -> {}\n", leaf_ref.id(), next_ref.id()))?;
-                }
             }
         }
 
@@ -842,7 +872,10 @@ where
 {
     fn drop(&mut self) {
         if !self.deprecate {
-            self.lock.set_tag(Tag::None, Release);
+            let unlocked = self
+                .lock
+                .update_tag_if(Tag::None, |t| t == Tag::First, Release);
+            debug_assert!(unlocked);
         }
     }
 }
