@@ -121,14 +121,19 @@ where
     }
 
     /// Removes an entry associated with the given key.
-    pub fn remove<Q>(&self, key: &Q, barrier: &Barrier) -> Result<bool, RemoveError>
+    pub fn remove_if<Q, F: FnMut(&V) -> bool>(
+        &self,
+        key: &Q,
+        condition: &mut F,
+        barrier: &Barrier,
+    ) -> Result<bool, RemoveError>
     where
         K: Borrow<Q>,
         Q: Ord + ?Sized,
     {
         match &self.entry {
-            NodeType::Internal(internal_node) => internal_node.remove(key, barrier),
-            NodeType::Leaf(leaf_node) => leaf_node.remove(key, barrier),
+            NodeType::Internal(internal_node) => internal_node.remove_if(key, condition, barrier),
+            NodeType::Leaf(leaf_node) => leaf_node.remove_if(key, condition, barrier),
         }
     }
 
@@ -190,11 +195,7 @@ where
     }
 
     /// Removes the current root node.
-    pub fn remove_root(
-        root: &AtomicArc<Node<K, V>>,
-        do_not_remove_valid_root: bool,
-        barrier: &Barrier,
-    ) -> bool {
+    pub fn remove_root(root: &AtomicArc<Node<K, V>>, barrier: &Barrier) -> bool {
         let mut root_ptr = root.load(Acquire, barrier);
         loop {
             if let Some(root_ref) = root_ptr.as_ref() {
@@ -217,10 +218,10 @@ where
                     root_ptr = root.load(Acquire, barrier);
                     continue;
                 }
-                if do_not_remove_valid_root && !root_ref.obsolete(barrier) {
+                if !root_ref.obsolete(barrier) {
                     break;
                 }
-                match root.compare_exchange(root_ptr, (None, Tag::None), Release, Relaxed) {
+                match root.compare_exchange(root_ptr, (None, Tag::None), Acquire, Acquire) {
                     Ok((old_root, _)) => {
                         // It is important to keep the locked state until the node is dropped
                         // in order for all the on-going structural changes being made to its
@@ -494,7 +495,7 @@ where
                 return Err(InsertError::Retry((key, value)));
             }
 
-            let unbounded_ptr = self.children.1.load(Relaxed, barrier);
+            let unbounded_ptr = self.children.1.load(Acquire, barrier);
             if let Some(unbounded_ref) = unbounded_ptr.as_ref() {
                 debug_assert!(unbounded_ptr.tag() == Tag::None);
                 if !(self.children.0).validate(result.1) {
@@ -528,7 +529,12 @@ where
     }
 
     /// Removes an entry associated with the given key.
-    fn remove<Q>(&self, key: &Q, barrier: &Barrier) -> Result<bool, RemoveError>
+    fn remove_if<Q, F: FnMut(&V) -> bool>(
+        &self,
+        key: &Q,
+        condition: &mut F,
+        barrier: &Barrier,
+    ) -> Result<bool, RemoveError>
     where
         K: Borrow<Q>,
         Q: Ord + ?Sized,
@@ -542,7 +548,7 @@ where
                     continue;
                 }
                 if let Some(child_ref) = child_ptr.as_ref() {
-                    return match child_ref.remove(key, barrier) {
+                    return match child_ref.remove_if(key, condition, barrier) {
                         Ok(removed) => Ok(removed),
                         Err(remove_error) => match remove_error {
                             RemoveError::Empty(removed) => self.coalesce(removed, barrier),
@@ -560,7 +566,7 @@ where
                     // Data race resolution - see LeafNode::search.
                     continue;
                 }
-                return match unbounded_ref.remove(key, barrier) {
+                return match unbounded_ref.remove_if(key, condition, barrier) {
                     Ok(removed) => Ok(removed),
                     Err(remove_error) => match remove_error {
                         RemoveError::Empty(removed) => self.coalesce(removed, barrier),
@@ -897,7 +903,7 @@ where
             let node_ptr = entry.1.load(Relaxed, barrier);
             let node_ref = node_ptr.as_ref().unwrap();
             if node_ref.obsolete(barrier) {
-                self.children.0.remove(entry.0);
+                self.children.0.remove_if(entry.0, &mut |_| true);
                 // Once the key is removed, it is safe to deallocate the node as the validation
                 // loop ensures the absence of readers.
                 if let Some(node) = entry.1.swap((None, Tag::None), Release) {
@@ -918,7 +924,7 @@ where
                         Release,
                     );
                     // Then, removes the node from the children list.
-                    if self.children.0.remove(max_entry.0).2 {
+                    if self.children.0.remove_if(max_entry.0, &mut |_| true).2 {
                         // Retires the children if it was the last child.
                         let result = self.children.0.retire();
                         debug_assert!(result);
